import copy
import logging
import os
import sys
from logging.handlers import TimedRotatingFileHandler

from sklearn.base import RegressorMixin

def add_filehandler(name='pytolemaic', log_path='default'):
    fmt = logging.Formatter("[%(asctime)s][%(name)s][%(levelname)s]: %(message)s")

    if log_path == 'default':
        abs_path = os.path.abspath(__file__)
        log_path = os.path.join(abs_path, '../../../logs/log.log')
    file_handler = TimedRotatingFileHandler(log_path, when='midnight')
    file_handler.setFormatter(fmt)
    file_handler.setLevel(logging.DEBUG)
    logger = get_logger(name)
    logger.addHandler(file_handler)
    return logger

def set_logger(name, level="DEBUG", handlers=()):
    logger = get_logger(name)
    if level:
        logger.setLevel(level=os.environ.get("LOGLEVEL", level))

    if handlers is not None:
        if len(handlers)==0:
            stream_handler = logging.StreamHandler(sys.stdout)
            fmt = logging.Formatter("[%(asctime)s][%(name)s][%(levelname)s]: %(message)s",
                                    datefmt='%Y-%m-%d %H:%M:%S')
            stream_handler.setFormatter(fmt)
            stream_handler.setLevel(max(logger.level, logging.INFO))

            handlers = [stream_handler]

        for h in handlers:
            logger.addHandler(h)

    return logger

def get_logger(name):
    return logging.getLogger(name)

    # split = name.split('.')
    # name = split[0]
    # if len(split)>2 and split[0]=='pytolemaic':
    #     name = split[0] + '.' + split[-1]

set_logger(name='pytolemaic')
logger = get_logger(__name__)

matplotlib_logger = logging.getLogger("matplotlib")
matplotlib_logger.setLevel(os.environ.get("LOGLEVEL", "WARNING"))




import numpy
import pandas
from matplotlib._color_data import XKCD_COLORS
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline

global tictoc
from time import time


def tic(name=None):
    global tictoc

    tictoc = time()
    if name is not None:
        logger.info("Calculating {}...".format(name))


def toc(name=None):
    global tictoc
    if name is not None:
        logger.info("Calculating {}... Done {:.2g} seconds".format(name, time() - tictoc))
    return time() - tictoc


class GeneralUtils():

    @classmethod
    def is_classification(cls, model):
        """
        Use the existence of predict_proba method to differentiate between classifier and regressor
        In order to use a regressor with a predict_proba method introduce an _estimator_type property with value "regressor"
        """
        is_regression = getattr(model, "_estimator_type", None) in ["regressor", RegressorMixin._estimator_type]
        return hasattr(model, 'predict_proba') and callable(model.predict_proba) and not is_regression

    @classmethod
    def dmd_supported(cls, model, dmd):
        try:
            model.predict(dmd.split_by_indices(indices=[0, 1, 2]))
            return True
        except:
            return False

    @classmethod
    def simple_imputation_pipeline(cls, estimator):
        # note - this works only for enncoded data (e.g. no text in caategoricl features)
        estimators = []
        estimators.append(('Imputer', SimpleImputer()))
        estimators.append(('Estimator', estimator))
        return Pipeline(steps=estimators)

    @classmethod
    def round_values(cls, d: dict, digits=5):
        for k, v in d.items():
            if isinstance(v, dict):
                cls.round_values(v, digits=digits)
            else:
                try:
                    d[k] = numpy.round(v, digits)
                except:
                    pass

        return d

    @classmethod
    def f5(cls, x):
        return numpy.round(x, 5)

    @classmethod
    def f3(cls, x):
        return numpy.round(x, 3)

    @classmethod
    def add_nans(cls, x, ratio=0.1):
        rs = numpy.random.RandomState(0)
        # let's add some missing values
        nan_locs = numpy.ones(numpy.prod(x.shape))
        nan_locs[rs.permutation(len(nan_locs))[:int(ratio * len(nan_locs))]] = numpy.nan
        nan_locs = nan_locs.reshape(x.shape)
        x = x * nan_locs
        return x

    @classmethod
    def shuffled_colors(cls):
        l = list(XKCD_COLORS.values())
        rs = numpy.random.RandomState(0)
        return rs.permutation(l)

    @classmethod
    def make_dict_json_compatible(cls, dictionary: dict):
        
        def retype(obj):
            if isinstance(obj, (numpy.int_, numpy.intc, numpy.intp, numpy.int8,
                                numpy.int16, numpy.int32, numpy.int64, numpy.uint8,
                                numpy.uint16, numpy.uint32, numpy.uint64)):
                return int(obj)
            elif isinstance(obj, (numpy.float_, numpy.float16, numpy.float32,
                                  numpy.float64)):
                return float(obj)
            elif isinstance(obj, (numpy.ndarray,)): #### This is the fix
                return obj.tolist()
            return  obj

        dictionary = copy.deepcopy(dictionary)
        for k, v in dictionary.items():
            if isinstance(v, dict):
                dictionary[k] = cls.make_dict_json_compatible(v)
            else:
                dictionary[k] = retype(v)

        return dictionary

    @classmethod
    def shorten_long_dict(cls, dictionary: dict, dict_limit=20, list_limit=10):
        if len(dictionary) > dict_limit:
            keys = list(dictionary.keys())[:dict_limit]
            dictionary = {key: dictionary[key] for key in keys}
            dictionary['...'] = '...'

        for k, v in dictionary.items():
            if isinstance(v, dict):
                dictionary[k] = cls.shorten_long_dict(v)
            elif isinstance(v, (list, tuple)):
                if len(v) > list_limit:
                    dictionary[k] = [v[0], v[1], '...', v[-2], v[-1]]
            else:
                pass

        return dictionary

    @classmethod
    def make_dict_printable(cls, dictionary: dict):

        dictionary = cls.shorten_long_dict(dictionary)
        dictionary = cls.round_values(dictionary, digits=5)
        dictionary = cls.make_dict_json_compatible(dictionary)

        return dictionary

    @classmethod
    def nan_mask(cls, x):
        return pandas.DataFrame(x).isnull().values



from typing import List

from matplotlib import pyplot as plt

import time

import gower
import numpy
from xgboost import XGBRegressor

from pytolemaic.utils.dmd import DMD
from pytolemaic.utils.constants import FeatureTypes
from pytolemaic.utils.metrics import Metrics, Metric
from sklearn.preprocessing import LabelEncoder

from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
class FeatureValueAnomalyDetector():

    def __init__(self, base_estimator='dt', use_target=True,
                 reg_metric='r2', clas_metric='recall',
                 clas_threshold=0.99,
                 reg_method = 'binning',
                 contamination= 0.02,
                 n_stds = None,
                 range_ratio = 0.05,
                 **submodel_kwargs):
        self.base_estimator = base_estimator

        self.submodel_kwargs = submodel_kwargs
        self.use_target = use_target
        self.models = {}

        self.reg_metric = reg_metric
        self.clas_metric = clas_metric
        self.clas_threshold = clas_threshold
        self.reg_method = reg_method
        self.contamination = contamination
        self.n_stds = n_stds
        self.range_ratio = range_ratio # to discard small anomalies when std is almost 0

    def _get_metric(self, metric_name)->Metric:
        metric = Metrics.supported_metrics().get(metric_name, None)
        if metric is None:
            raise NotImplementedError("metric {} is no implemented".format(metric_name))

        return metric

    def get_model(self, is_classification):
        from xgboost import XGBRegressor, XGBClassifier
        kwargs = dict(random_state=0,
                       n_estimators=250,
                       n_jobs=-1)
        kwargs.update(self.submodel_kwargs)
        model = XGBClassifier(**kwargs) if is_classification else XGBRegressor(**kwargs)
        return model

    def analyze(self, dmd_train: DMD, features_to_analyze:List[str]=None, perform_extra_iteration=False):
        inds1, inds2, train1, train2 = self.split_train_into_2_parts(dmd_train)

        if features_to_analyze:
            features_subset_ = [i for i in range(dmd_train.n_features)
                                   if dmd_train.feature_names[i] in features_to_analyze]
            if len(features_subset_)==0:
                raise KeyError("{} not recognized".format(features_to_analyze))
            else:
                features_to_analyze = features_subset_
        else:
            features_to_analyze = range(dmd_train.n_features)

        reports = {}
        for ifeature in features_to_analyze:
            name = dmd_train.feature_names[ifeature]
            print("\n\nAnalyzing feature #{}: {}".format(ifeature, name))

            report1, report2 = self.calculate_anomaly_scores(
                 train1=train1, train2=train2, ifeature=ifeature,
                 feature_types=dmd_train.feature_types, perform_extra_iteration=perform_extra_iteration)

            for report in [report1, report2]:
                report['anomaly_score_ratio'] = report['anomaly_score']/report['threshold']


            report = {}
            for k in report1:
                v1 = report1[k]
                v2 = report2[k]
                if isinstance(v1, list):
                    v1 = numpy.array(v1)
                    v2 = numpy.array(v2)
                if isinstance(v1, numpy.ndarray):
                    if len(v1.shape)==1:
                        v = numpy.zeros(dmd_train.n_samples) #1D
                    else:
                        v = numpy.zeros((dmd_train.n_samples, v1.shape[1])) #2D
                    v[inds1] = v1
                    v[inds2] = v2
                    report[k] = v
                elif v1 == v2:
                    report[k] = v1
                else:
                    report[k] = (v1+v2)/2

            v = report['anomaly_score_ratio']
            print("Found {} samples with anomaly score >= {:.3g}, and {} (out of {}) samples with anomaly score > 0".format(
                sum(v>=1), report['threshold'], sum(v > 0), len(v)))

            reports[name] = report
        return reports
        #
        #     anomaly_score_raw[inds1, ifeature] = report1['anomaly_score']
        #     anomaly_score_raw[inds2, ifeature] = report2['anomaly_score']
        #
        #     anomaly_score_threshold[inds1, ifeature] = report1['anomaly_score']/report1['threshold']
        #     anomaly_score_threshold[inds2, ifeature] = report2['anomaly_score']/report2['threshold']
        #
        #     yt[inds1, ifeature] = report1['yt']
        #     yt[inds2, ifeature] = report2['yt']
        #
        #     yp[inds1, ifeature] = report1['yp']
        #     yp[inds2, ifeature] = report2['yp']
        #
        #
        #     v = anomaly_score_threshold[:, ifeature]
        #     print("Found {} samples with anomaly score >= {:.3g}, and {} (out of {}) samples with anomaly score > 0".format(
        #         sum(v>=1), min(report1['threshold'], report2['threshold']), sum(v > 0), len(v)))
        #
        #     if False:
        #         sorted_inds[name] = sorted(numpy.arange(len(v)), key=lambda j: v[j], reverse=True)
        #         v = anomaly_score_raw[:,ifeature][sorted_inds[name]]
        #         plt.plot(v, '.')
        #         label = "Threshold for contamination = {}%".format(self.contamination*100) \
        #             if self.contamination else "Threshold for {} Sigma".format(self.n_stds)
        #         plt.plot([0,len(v)],[threshold1, threshold1], '-k', label=label)
        #         plt.plot([0,len(v)],[threshold2, threshold2], '-k')
        #         plt.show()
        #
        # if True:
        #     inds = numpy.arange(len(anomaly_score_threshold))
        #     sm = numpy.sum(anomaly_score_threshold>=1,axis=1)
        #     inds = sorted(inds, key=lambda k:sm[k])
        #     for ifeature in range(0, dmd_train.n_features):
        #         v = anomaly_score_threshold[inds, ifeature].copy()
        #         v[v<1]=0
        #
        #
        #         print(ifeature, dmd_train.feature_names[ifeature], sum(v), sum(v)/len(v))
        #         plt.plot(v, '-', label='{}'.format(dmd_train.feature_names[ifeature]))
        #     plt.legend()
        #     plt.show()
        # if True:
        #     for ifeature in range(0, dmd_train.n_features):
        #         v = anomaly_score_threshold[:, ifeature].copy()
        #
        #         if numpy.max(v)<1:
        #             continue
        #
        #         fig, axs = plt.subplots(2,1, sharex=True)
        #         inds = numpy.arange(len(v))[v>=1]
        #         axs[0].plot(inds,anomaly_score_raw[v>=1, ifeature],'.', label=dmd_train.feature_names[ifeature])
        #         axs[1].plot(inds,yt[v>=1, ifeature],'.b', label='Data')
        #         axs[1].plot(inds,yp[v>=1, ifeature],'.r', label='Expected')
        #         plt.legend()
        #         plt.show()
        #
        # return anomaly_score_raw

    def calculate_anomaly_scores(self, train1, train2, feature_types, ifeature, perform_extra_iteration):
        print("Analysis started")

        report2 = self._analyze_one_way(ifeature, train=train1, validation=train2,
                                        feature_types=feature_types, metric_score_only=perform_extra_iteration)
        report1 = self._analyze_one_way(ifeature, train=train2, validation=train1,
                                        feature_types=feature_types)
        metric = report1['metric']
        print("P1 train, P2 test ({}): {:.3g}".format(metric, report2['metric_score']))
        print("P2 train, P1 test ({}): {:.3g}".format(metric, report1['metric_score']))

        if perform_extra_iteration: # clean the train data from anomalies and re-iterate
            # 1st iteration
            inds_to_keep1 = report1['anomaly_score'] < report1['threshold']
            report2 = self._analyze_one_way(ifeature=ifeature, train=train1[inds_to_keep1, :], validation=train2,
                                            feature_types=feature_types)

            # 2nd iteration
            inds_to_keep2 = report2['anomaly_score'] < report2['threshold']
            report1 = self._analyze_one_way(ifeature=ifeature, train=train2[inds_to_keep2, :], validation=train1,
                                            feature_types=feature_types)

            print("Cleaned P1 train, P2 test ({}): {:.3g}".format(metric, report2['metric_score']))
            print("Cleaned P2 train, P1 test ({}): {:.3g}".format(metric, report1['metric_score']))
        return report1, report2

    def split_train_into_2_parts(self, dmd_train):
        rs = numpy.random.RandomState(0)
        inds = rs.permutation(dmd_train.n_samples)
        inds1, inds2 = inds[:len(inds) // 2], inds[len(inds) // 2:]
        train1, train2 = dmd_train.split_by_indices(inds1), dmd_train.split_by_indices(inds2)
        if self.use_target and dmd_train.target is not None:
            train1 = numpy.concatenate([train1.values, train1.target], axis=1)
            train2 = numpy.concatenate([train2.values, train2.target], axis=1)
        else:
            train1 = train1.values
            train2 = train2.values
        train1 = train1.astype(float)
        train2 = train2.astype(float)
        return inds1, inds2, train1, train2

    def _get_xy(self, data: numpy.ndarray, itarget, drop_na):
        features_wo_i = list(range(data.shape[1]))
        features_wo_i.remove(itarget)

        if drop_na:
            isfinite = numpy.isfinite(data[:, itarget])
            data = data[isfinite, :]

        x = data[:, features_wo_i].copy()
        y = data[:, itarget].copy()

        return x, y

    def _encode_target(self, target, is_classification, le: LabelEncoder=None, bins=None):

        if not is_classification and self.reg_method=='binning':
            if bins is None:
                bins = self.reg_binning(target)

            target = numpy.digitize(numpy.clip(target, bins[0], bins[-1]),
                                    bins=bins)

        if is_classification or self.reg_method == 'binning':
            target = target.astype(int)
            if le is None:
                le = LabelEncoder()
                le.fit(target)
            try:
                target = le.transform(target).astype(int)
            except:
                target = numpy.array([le.transform([y_]).astype(int)[0]
                                  if y_ in le.classes_ and not numpy.isnan(y_)
                                  else numpy.nan
                                  for y_ in target])

        return target, bins, le

    def reg_binning(self, y, max_bins=20):
        u = numpy.unique(y)
        if numpy.nanmax(u-u.astype(int))==0 and len(u)<=max_bins:
            bins = u
        else:
            bins = numpy.histogram_bin_edges(y, bins='auto')  # todo: fix range

        y_ = numpy.digitize(y, bins=bins)
        u, c = numpy.unique(y_, return_counts=True)
        if len(bins) != max(y_) or len(bins) > max_bins or min(c) < 20:
            # empty bins
            print("len(bin)=={}, max(y)=={}".format(len(bins), max(y)))

            bins = numpy.percentile(y, numpy.arange(max_bins) * 100/max_bins)
            bins = numpy.histogram_bin_edges(y, bins=bins)
            bins = numpy.unique(bins)
            print("len(bin)=={}, max(y)=={}".format(len(bins), max(y)))
        return bins


    def _calculate_score(self, yv, yp, is_classification):
        metric_name = self.clas_metric if is_classification else self.reg_metric

        metric = self._get_metric(metric_name=metric_name)
        if metric.is_proba:
            raise NotImplementedError("Proba metrics are not supported yet")

        isfinite = numpy.isfinite(yv)
        score = metric.function(yv[isfinite], yp[isfinite])
        score = Metrics.metric_as_score(score, metric)
        return score

    # def _analyze_yproba_highpass(self, yv, yp, yproba):
    #     max_yproba = numpy.max(yproba, axis=1)
    #     # do not filter yv==nan to maintain correct indexing
    #     isfinite = numpy.isfinite(yv)
    #
    #     p_out = None
    #     for j in reversed(range(10)):  # p: 0.7 --> 0.97
    #         p = 1 - (j + 1) / 30
    #         inds = max_yproba >= p
    #         correct = yv[inds] == yp[inds]  # yv[j]==nan --> correct[i]=False
    #         if numpy.sum(correct) / numpy.sum(inds & isfinite) > self.clas_threshold:
    #             p_out = p
    #             break
    #
    #     if p_out is None:
    #         print("highpass:clas_threshold {} not met".format(self.clas_threshold))
    #         return []
    #     else:
    #
    #         high_pass = numpy.arange(len(yv))[max_yproba > p_out]
    #         out = [j for j in high_pass if yp[j] != yv[j] and isfinite[j]]
    #         print("highpass: clas_threshold with threshold {:.3g} - len(out)={}".format(p_out, len(out)))
    #
    #         return out

    def _analyze_yproba(self, yv, yproba, key='low'):
        """
        key=='low': model is very confident that the correct answer is incorrect:
            # locate candidates which are incorrect (candidates != max_yproba)
            # and the model is certain they are incorrect (candidates <= p_out)
            # for these samples, anomaly score is 1 - candidates
        key=='high': model is very confident in a wrong answer is correct:
                # locate candidates which are incorrect (candidates != max_yproba)
                # and the model is certain that some other class is correct (max_yproba >= p_out)
                # for these samples, anomaly score is candidates

        * candidates == y_proba for the expected class
        """
        isfinite = numpy.isfinite(yv)

        candidates = numpy.zeros_like(yv, dtype=float)
        candidates[~isfinite] = numpy.nan
        # yproba[i, int(yv[i])] is the probability for the correct class for sample 'i'
        candidates[isfinite] = [yproba[i, int(yv[i])] for i in numpy.arange(len(yv))[isfinite]]

        max_yproba = numpy.max(yproba, axis=1)

        # find a threshold p for which almost all candidates are > p
        p_out = None
        for j in reversed(range(11)): # Note order is reversed
            if key == 'low':
                p = j / 30 # gradually lower p
                correct = candidates > p  # candidate==nan --> cond=False. '>' because p can be 0
                denominator = numpy.sum((max_yproba>p) & isfinite) # how many samples has max_proba > p
            else:
                p = 1 - j/30 # gradually increase p
                correct = candidates >= p  # candidate==nan --> cond=False. '>=' because p can be 1
                denominator = numpy.sum((max_yproba>=p) & isfinite) # how many samples has max_proba >= p

            # print(key, j, p, numpy.sum(correct), denominator, numpy.sum(correct) / denominator)

            # condition: at least some samples has max_proba > p, and enough candidates are correct:
            if denominator > 0 and numpy.sum(correct) / denominator > self.clas_threshold:
                p_out = p
                break
            else:
                continue

        out = numpy.zeros_like(candidates, dtype=float)
        if p_out is None:
            print("{}pass:clas_threshold {} not met".format(key, self.clas_threshold))
            return out
        else:
            if key=='low':
                # model is very confident (candidates <= p_out) that the expected class is not the correct answer (candidates != max_yproba)
                incorrect = (candidates <= p_out) & (candidates != max_yproba) & isfinite
            else:
                # model is very confident (max_yproba >= p_out) that the correct answer is not the expected class (candidates != max_yproba)
                incorrect = (max_yproba >= p_out) & (candidates != max_yproba) & isfinite

            out[incorrect] = 1-candidates[incorrect]
            return out

    def _analyze_one_way(self, ifeature, train: numpy.ndarray, validation: numpy.ndarray,
                         feature_types, metric_score_only=False):

        """
        report_regression = \
            dict(yt=[],
                 yp=[],
                 n_stds = [],
                 yrange=None,
                 anomaly_score=[],
                 threshold=None,
                 metric_score=None)
        report_classification = \
            dict(yt=[],
                 yp=[],
                 probabiliy_of_yt = [], # y_proba of correct class
                 probabiliy_of_yp = [], # y_proba of highest proba
                 anomaly_score=[],
                 threshold=None,
                 metric_score=None)
        """

        is_classification = feature_types[ifeature] == FeatureTypes.categorical

        ## extract data. if reg is 'binning' convert to a classification task
        x, y = self._get_xy(data=train, itarget=ifeature, drop_na=True)
        xv, yv = self._get_xy(data=validation, itarget=ifeature, drop_na=False)


        if is_classification or self.reg_method == 'binning':
            if is_classification: # code to avoid 1 class in 1 set, but not in the other
                cond = numpy.in1d(y, yv)
                cond[~numpy.isfinite(y)] = True # ignore results for nans
                x = x[cond,:] # drop all instances which are not in yv
                y = y[cond] # drop all instances which are not in yv

                cond = numpy.in1d(yv, y)
                cond[~numpy.isfinite(yv)] = True  # ignore results for nans
                yv[~cond]=numpy.nan  # drop all instances which are not in y

            y, bins, le = self._encode_target(y, is_classification=is_classification, le=None, bins=None)
            yv, _, _ = self._encode_target(yv, is_classification=is_classification, le=le, bins=bins)
            is_classification = True

        model = self.get_model(is_classification=is_classification) # e.g. XGBoost
        model.fit(x, y)
        yp = model.predict(xv).astype(float)

        # average error (score) for entire validation set
        isfinite = numpy.isfinite(yv)
        metric = self.clas_metric if is_classification else self.reg_metric
        metric_score = self._calculate_score(yv[isfinite], yp[isfinite], is_classification=is_classification)
        if metric_score_only:
            return dict(metric_score=metric_score, metric=metric)

        if is_classification:
            yproba = model.predict_proba(xv)


            # print('highpass - certainty in incorrect answer')
            out_high = self._analyze_yproba(yv, yproba, key='high')
            # print('lowpass - certainty the correct answer is incorrect')
            out_low = self._analyze_yproba(yv, yproba, key='low')

            out_anomaly_score = numpy.max(
                numpy.concatenate([out_low.reshape(1,-1), out_high.reshape(1,-1)]),
                axis=0)

            # plt.plot(out_score,'.')
            # plt.show()

            threshold = numpy.percentile(out_anomaly_score, 100 * (1 - self.contamination))
            if threshold == numpy.min(out_anomaly_score):
                threshold = numpy.min(out_anomaly_score)+1
            threshold = max(threshold, 0.75)

            yp[~numpy.isfinite(yv)] = numpy.nan
            report_classification = \
                dict(yt=yv,
                     yp=yp,
                     probabiliy_of_yt=[yproba[i, int(yv[i])] if numpy.isfinite(yv[i]) else numpy.nan for i in numpy.arange(len(yv))],  # y_proba of correct class
                     probabiliy_of_yp=yproba,  # y_proba of highest proba
                     anomaly_score=out_anomaly_score,
                     threshold=threshold,
                     metric_score=metric_score,
                     metric=metric)

            return report_classification #report_classification['metric_score'], report_classification['anomaly_score'], report_classification['threshold']
        else:
            scale = lambda d: (d - numpy.nanmean(d))/(numpy.nanstd(d)+1e-10)
            yrange = numpy.max(y) - numpy.min(y)


            deltas = yv - yp
            deltas_orig = deltas.copy()
            deltas_scaled = scale(deltas)

            if 'dt' in self.reg_method or 'xgb' in self.reg_method:
                # trying to divide the space into subspaces with similar error.
                y_delta = deltas_scaled[isfinite]
                noise = (1 + 1e-3 * rs.rand(len(y_delta)))
                y_delta *= noise

                x_ = xv[isfinite,:].copy()

                proximity_model = XGBRegressor(min_child_weight=100,
                                               n_estimators=1,
                                               # num_feature=xv.shape[1],
                                               subsample=1,
                                               colsample_bynode=1,
                                               # reg_lambda=0,
                                               tree_method='exact',
                                               max_depth=100,
                                               random_state=iter)
                # proximity_model = DecisionTreeRegressor(min_samples_leaf=100)
                proximity_model.fit(x_, y_delta)  # y=yv or y=delta?

                y_proximity = proximity_model.predict(xv)
                u,c = numpy.unique(y_proximity, return_counts=True)

                out_score = numpy.zeros(len(xv))
                n_stds = numpy.zeros(len(xv))*numpy.nan
                for leaf_ind, value in enumerate(u):
                    indices = numpy.argwhere((y_proximity==value) & isfinite).ravel()
                    if len(indices)==0:
                        continue

                    deltas_of_leaf_orig = deltas_orig[indices]

                    n_stds[indices] = numpy.abs(scale(deltas_of_leaf_orig))

                # score is 1-1/#std. e.g. std=2 --> score=0.5. std=3 --> score=0.66
                # high score --> high anomaly
                anomaly_score = 1 - 1 / n_stds

                # points with low error are considered normal, even if out of distribution for leaf
                anomaly_score[numpy.abs(yp-yv) < self.range_ratio * yrange] = 0
                anomaly_score = numpy.clip(anomaly_score, 0, 1)

                if self.n_stds:
                    threshold = 1 - 1 / self.n_stds
                else:
                    threshold = numpy.percentile(anomaly_score, 100 * (1 - self.contamination))
                    if threshold == numpy.min(anomaly_score):
                        threshold = numpy.min(anomaly_score) + 1

                yp[~numpy.isfinite(yv)] = numpy.nan
                report_regression = \
                    dict(yt=yv,
                         yp=yp,
                         n_stds=n_stds,
                         yrange=yrange,
                         anomaly_score=anomaly_score,
                         threshold=threshold,
                         metric_score=metric_score,
                         metric=metric)

                return report_regression #metric_score, out_score, threshold
            else:
                raise NotImplementedError(str(self.reg_method))



if __name__ == '__main__':

    from resources.datasets.california_housing import CaliforniaHousing
    from resources.datasets.uci_adult import UCIAdult
    from resources.datasets.linear import LinearRegressionDataset
    import os, pandas
    from pytolemaic import FeatureTypes, DMD, HOME_DIR, PyTrust
    num, cat = FeatureTypes.numerical, FeatureTypes.categorical

    # # train, test = CaliforniaHousing().as_dmd()
    # train, test = UCIAdult().as_dmd()
    # train = DMD.concat([train,test])
    #
    # ##

    #
    train_path = os.path.join(HOME_DIR, 'resources', 'datasets', 'titanic-train.csv')
    df_train = pandas.read_csv(train_path)
    feature_types = [num, cat, cat, cat, num, num, num, cat, num, cat, cat]
    feature_types = [t for i,t in enumerate(feature_types) if df_train.columns[i] not in ['PassengerId', 'Pclass', 'Name']]
    df_train.drop(columns=['PassengerId', 'Pclass', 'Name'], inplace=True)

    dmd_train, dmd_test = DMD.from_df(df_train=df_train, df_test=None,
                                      is_classification=True,
                                      target_name='Survived',
                                      feature_types=feature_types,
                                      categorical_encoding=True, nan_list=['?'],
                                      split_ratio=None)

    dmd_train,dmd_test = UCIAdult().as_dmd()
    print(dmd_train.feature_names)

    ##
    """"""
    rs = numpy.random.RandomState(0)
    x = rs.rand(100000,3)
    x[:,0] = x[:,1] + x[:,2]  + 1e-2*rs.rand(len(x))
    x[:12,0]+=0.2
    # x[10:20,0]*=0.2
    y = rs.rand(100000)
    #
    # dmd_train = DMD(x, y,
    #                 feature_types=[num]*x.shape[1],
    #                 )

    fad = FeatureValueAnomalyDetector(
        base_estimator='xgb', use_target=False,
                 reg_metric='r2', clas_metric='recall',
                 clas_threshold=0.95,
                contamination=0.001,
                n_stds=3,
                # reg_method='binning',
                reg_method='xgb',
        eval_metric='merror', use_label_encoder=False, # xgb kwargs
    )
    out_report = fad.analyze(dmd_train, features_to_analyze=['sex'])

    from anomaly_values_in_data_report import AnomaliesInDataReport
    report = AnomaliesInDataReport(out_report, dmd_train.categorical_encoding_by_feature_name)

    print("\n".join(report.insights(n_top_features=3)))

    report.plot(plot_only_above_threshold=False)
    plt.show()



import numpy
from matplotlib import pyplot as plt
from sklearn.utils.multiclass import unique_labels

from pytolemaic.utils.base_report import Report

from pytolemaic.utils.general import get_logger

logger = get_logger(__name__)

class AnomaliesInDataReport(Report):

    def __init__(self, anomalies_report : dict, categorical_encoding_by_feature_name):
        self.categorical_encoding = categorical_encoding_by_feature_name
        supported_keys = self.to_dict_meaning().keys()
        self._report = {}
        for feature_name, feature_report in anomalies_report.items():
            if len(set(feature_report.keys()) - set(supported_keys)):
                # excess of keys
                msg = ""
                for key, value in feature_report.items():
                    if key not in supported_keys:
                       msg += "key={} (value={}) is not supported.\n".format(key, value)
                raise NotImplementedError("Some keys are not supported:\n"+msg)

            self._report[feature_name] = {key: feature_report.get(key, None) for key in supported_keys}

    def to_dict(self, printable=False):
        return self._printable_dict(self._report, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(
            yt="Value that appear in data",
            yp="Expected value (model prediction)",
            probabiliy_of_yt="y_proba of correct class (classification)",
            probabiliy_of_yp="y_proba (classification)",
            n_stds = "Deviation from mean as # of stds (regression)",
            yrange="Range of feature: max(y) - min(y) (regression)",
            anomaly_score='Anomaly score. "1-1/n_stds" for numerical and "y_proba" for categorical',
            anomaly_score_ratio='Anomaly score divided by threshold',
            threshold='Threshold chosen/calculated, depending on analysis parameters',
            metric_score="Model score of model trained on 50% of data and tested on the other 50%",
            metric="Metric used to calculate metric_score")

    def plot(self, features_to_plot=None,
             plot_only_above_threshold=True,
             figsize=(10,10)):
        features_to_plot = features_to_plot or sorted(self._report.keys())

        for feature in features_to_plot:
            if feature not in self._report:
                raise KeyError("{} not recognized".format(feature))

            report = self._report[feature]


            v = report['anomaly_score_ratio']
            if numpy.all(v<1): # skip
                logger.info('Feature "{}" hsa no anomalies, skipping'.format(feature))
                continue

            if report['n_stds'] is not None:
                self._plot_regression(feature_name=feature, report=report, plot_only_above_threshold=plot_only_above_threshold)
            else:
                self._plot_classification(feature_name=feature, report=report, plot_only_above_threshold=plot_only_above_threshold)

            # plt.show()

    def _plot_regression(self, feature_name, report, plot_only_above_threshold):
        # 1 scatter
        ax = plt.figure(figsize=(10,10)).add_subplot()

        is_finite = numpy.isfinite(report['yt'])
        if plot_only_above_threshold:
            subset = report['anomaly_score_ratio'] >= 1
        else:
            subset = report['anomaly_score_ratio'] >= 0
        subset = subset & is_finite

        x,y = report['yt'][subset], report['yp'][subset]
        ax.scatter(x,y, c=report['anomaly_score_ratio'][subset])
        ax.plot([numpy.min(x), numpy.max(x)], [numpy.min(x), numpy.max(x)],':')
        ax.set_xlabel('"{}": Value in data'.format(feature_name))
        ax.set_ylabel('"{}": Expected value'.format(feature_name))
        ax.set_title('Scatter plot for actual and expected values for feature "{}"'.format(feature_name))

        plt.tight_layout()

        # plt.show()

    def _plot_classification(self, feature_name, report, plot_only_above_threshold):
        from pytolemaic.analysis_logic.model_analysis.scoring.scoring_report import ConfusionMatrixReport
        from sklearn.metrics import confusion_matrix

        fig, (ax1, ax2) = plt.subplots(1,2,figsize=(15, 10))

        is_finite = numpy.isfinite(report['yt'])

        for ax in [ax1, ax2]: #ax3
            # ax3 is used to weight confusion matrix samples by anomaly score. Doesn't seem to be interesting...
            yt = report['yt'][is_finite]
            yp = report['yp'][is_finite]
            anomaly_score_ratio = report['anomaly_score_ratio'][is_finite]
            cm_labels = unique_labels(yt, yp)

            if ax==ax1:
                title='Confusion matrix ("{}")'.format(feature_name)
                weights = None
            else:
                subset = anomaly_score_ratio >= 1
                yt, yp = yt[subset], yp[subset]
                # if plot_only_above_threshold:
                #     cm_labels = unique_labels(yt, yp)

                if ax==ax2:
                    title = 'Confusion matrix for anomalous values ("{}")'.format(feature_name)
                    weights = None
                else:
                    title = 'Confusion matrix for anomalous values weighted by anomaly score ("{}")'.format(feature_name)
                    weights = anomaly_score_ratio[subset]

            cm = confusion_matrix(y_true=yt, y_pred=yp, labels=cm_labels,
                                  sample_weight=weights).tolist()

            ConfusionMatrixReport._plot_confusion_matrix(
                confusion_matrix=cm,
                labels=[self.categorical_encoding[feature_name][k] for k in cm_labels],
                title=title,
                ax=ax)

        plt.tight_layout()
        # plt.show()

    def insights(self, n_top_features=5):
        insights_out = []
        n_anomalous_samples = {}
        scores = {}
        features = self._report.keys()
        for feature in features:
            report = self._report[feature]
            v = report['anomaly_score_ratio']
            # if numpy.all(v < 1):  # skip
            #     continue
            n_anomalous_samples[feature] = numpy.sum(v>=1)
            scores[feature] = report['metric_score']

        features = sorted(features, key=lambda f: n_anomalous_samples[f], reverse=True)
        if n_anomalous_samples[features[0]]>0:
            features = [f for f in features if n_anomalous_samples[f]>0]

            n_anomalous_samples_ = [n_anomalous_samples[f] for f in features]
            n_features_with_anomaly = numpy.sum(numpy.array(n_anomalous_samples_)>0)
            insight = "Found {} features with anomalous samples.".format(n_features_with_anomaly)
            insights_out.append(insight)


            insight = "Features with most anomalies are {} with {} anomalies respectively"\
                .format(features[:n_top_features], n_anomalous_samples_[:n_top_features])
            insights_out.append(insight)

            features = sorted(features, key=lambda f: scores[f], reverse=True)
            scores = [scores[f] for f in features]
            n_anomalous_samples_ = [n_anomalous_samples[f] for f in features]

            insight = "Features with highest score are {} with {}/{} score / # anomalies respectively" \
                .format(features[:n_top_features], scores[:n_top_features], n_anomalous_samples_[:n_top_features])
            insights_out.append(insight)

        return insights_out

"""

        as_str = lambda k: [str(s) for s in k]
        need_plot_for_features = any([len(v) > 0 for v in self.nan_counts_features.values()])
        need_plot_for_samples = any([len(v) > 0 for v in self.nan_counts_samples.values()])
        n_plots = int(need_plot_for_features) + int(need_plot_for_samples)
        if n_plots == 0:
            return

        # list of tuples
        f11 = sorted(self.nan_counts_features.items(), key=lambda kv: kv[0])  # sort by nan_ratio
        f21 = sorted(self.nan_counts_samples.items(), key=lambda kv: kv[0])  # sort by nan_ratio

        # list of tuples
        f12 = sorted(self.nan_counts_features[f11[0][0]].items(), key=lambda kv: kv[1], reverse=True)
        f22 = sorted(self.nan_counts_samples[f21[0][0]].items(), key=lambda kv: kv[1], reverse=True)
        f12 = f12[:min(len(f12), n_features_to_plot)]
        f22 = f22[:min(len(f22), n_features_to_plot)]

        if n_plots == 1:
            fig1, axs = plt.subplots(1, 2, figsize=(10, 5))

            ax11, ax12, ax21, ax22 = [None] * 4
            if need_plot_for_features:
                ax11, ax12 = axs
            else:
                ax21, ax22 = axs
        else:
            fig1, (ax11, ax12) = plt.subplots(1, 2, figsize=(10, 5))
            fig2, (ax21, ax22) = plt.subplots(1, 2, figsize=(10, 5))

        if ax11 is not None:
            keys, values = zip(*f11)
            x = list(reversed(keys))
            y = list(reversed([len(v) for v in values]))
            ax11.bar(x, y, width=0.2)
            ax11.set(
                title='Features with missing values above threshold',
                xlabel='Missing values threshold ',
                ylabel='Number of features',
                xticks=x)

            keys, values = zip(*f12)
            ax12.barh(as_str(reversed(keys)), list(reversed(values)))
            ax12.set(
                title='Features with highest missing ratio',
                xlabel='Missing values ratio',
                yticks=as_str(reversed(keys)))

        if ax22 is not None:
            keys, values = zip(*f21)
            x = list(reversed(keys))
            y = list(reversed([len(v) for v in values]))
            ax21.bar(x, y, width=0.2)
            ax21.set(
                title='Samples with missing values above threshold',
                xlabel='Missing values threshold ',
                ylabel='Number of samples',
                xticks=x)

            keys, values = zip(*f22)
            ax22.barh(as_str(reversed(keys)), list(reversed(values)))
            ax22.set(
                title='Samples with highest missing ratio',
                xlabel='Missing values ratio',
                yticks=as_str(reversed(keys)))

        plt.tight_layout()
        plt.draw()

    def insights(self):
        insights = []
        info = self.to_dict(printable=True)

        lvl1 = 0.1
        lvl2 = 0.5
        lvl3 = 0.9

        counts = [info['per_feature'][key]['count'] for key in info['per_feature']]
        if len(counts) > 0 and max(counts) > 0:
            max_key_features = max([key for key, values in info['per_feature'].items() if values['count'] > 0])
            missing_values = info['per_feature'][max_key_features]
            threshold = missing_values['threshold']

            sentence = "missing values ({}) were found in {} features".format(max_key_features, missing_values['count'])
            if threshold < lvl1:
                insights.append("Some {}.".format(sentence))
            elif threshold < lvl2:
                insights.append("Significant number of {}.".format(sentence))
            elif threshold < lvl3:
                insights.append(
                    "High number of {}. Check out these features:\n{}".format(sentence, missing_values['features']))
            else:
                insights.append("Extremely high number of {}. Check out these features:\n{}".format(sentence,
                                                                                                    missing_values[
                                                                                                        'features']))

        lvl1 = 0.1
        lvl2 = 0.5
        lvl3 = 0.9

        counts = [info['per_sample'][key]['count'] for key in info['per_sample']]
        if len(counts) > 0 and max(counts) > 0:
            max_key_samples = max([key for key, values in info['per_sample'].items() if values['count'] > 0])
            missing_values = info['per_sample'][max_key_samples]
            threshold = missing_values['threshold']

            sentence = "missing values ({}) were found in {} samples".format(max_key_samples, missing_values['count'])
            if threshold < lvl1:
                insights.append("Some {}.".format(sentence))
            elif threshold < lvl2:
                insights.append("Significant number of {}.".format(sentence))
            elif threshold < lvl3:
                insights.append(
                    "High number of {}. Check out these samples:\n{}".format(sentence, missing_values['samples']))
            else:
                insights.append("Extremely high number of {}. Check out these samples:\n{}".format(sentence,
                                                                                                   missing_values[
                                                                                                       'samples']))

        return self._add_cls_name_prefix(insights)


class DatasetAnalysisReport(Report):
    def __init__(self, class_counts, outliers_count, missing_values_report: MissingValuesReport,
                 covariance_shift_report:CovarianceShiftReport=None):
        self._class_counts = class_counts
        self._outliers_count = outliers_count
        self._missing_values_report = missing_values_report
        self._covariance_shift_report = covariance_shift_report


    @property
    def class_counts(self) -> dict:
        return self._class_counts

    @property
    def outliers_count(self) -> dict:
        return self._outliers_count

    @property
    def missing_values_report(self)->MissingValuesReport:
        return self._missing_values_report

    @property
    def covariance_shift_report(self)->CovarianceShiftReport:
        return self._covariance_shift_report

    def to_dict(self, printable=False):
        out = dict(few_class_representatives=self.class_counts,
                   outliers_count=self.outliers_count,
                   missing_values=self.missing_values_report.to_dict(printable=printable),
                   covariance_shift=None,
                   )
        if self.covariance_shift_report is not None:
            out.update(dict(covariance_shift=self.covariance_shift_report.to_dict(printable=printable)))
        return self._printable_dict(out, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(
            few_class_representatives='{feature_name: {class: instances_count}} : listing categorical features that has at least 1 class which is under-represented (less than 10 instances).',
            outliers_count='{feature_name: {n-sigma: outliers_info}} : listing numerical features that has more outliers than is expected with respect to n-sigma. '
                           'E.g. for 3-sigma we expect ceil(0.0027 x n_samples) outliers.',
            missing_values=MissingValuesReport.to_dict_meaning(),
            covariance_shift=CovarianceShiftReport.to_dict_meaning(),
        )

    def _plot_class_counts(self):
        if len(self.class_counts) == 0:
            return

        plt.figure()
        ax = plt.subplot()
        features = sorted(self.class_counts.keys())
        n_low_count_classes = [len(self.class_counts[feature]) for feature in features]

        x = [str(k) for k in reversed(features)]
        y = list(reversed(n_low_count_classes))
        plt.barh(x, y)
        ax.set(
            title='Features with under represented classes',
            xlabel='Number of under represented classes',
            ylabel='Features',
            yticks=x)

    def _plot_outlier_counts(self):
        if len(self.outliers_count) == 0:
            return

        plt.figure()
        ax = plt.subplot()
        sigmas = set(list(itertools.chain(*[v.keys() for v in self.outliers_count.values()])))

        features = sorted([str(k) for k in self.outliers_count.keys()], reverse=True)
        for sigma in sorted(sigmas):
            fv = [(feature, values[sigma]['n_outliers']) for feature, values in self.outliers_count.items() if
                  sigma in values]

            x, y = zip(*fv)

            x = [str(k) for k in list(reversed(x))]
            y = list(reversed(y))
            plt.barh(x, y)
        ax.set(
            title='Features with outliers',
            xlabel='Number of outliers',
            ylabel='Features',
            yticks=features)
        ax.legend(sorted(sigmas))

    def plot(self):
        self.missing_values_report.plot()
        self._plot_class_counts()
        self._plot_outlier_counts()
        if self.covariance_shift_report is not None:
            self.covariance_shift_report.plot()

    def _class_count_insights(self):
        insights = []
        for feature_name, class_counts in self.class_counts.items():
            if len(class_counts) == 1:
                sentence = "Feature '{}' contains a class with few representatives".format(feature_name)
            else:
                sentence = "Feature '{}' contains {} classes with few representatives - {} samples in total. " \
                           "With so few samples in each class there would be little to no learning.\n\tConsider merging all classes into a single 'other' class" \
                    .format(feature_name, len(class_counts), sum(class_counts.values()))

            list_of_classes = ["class '{}' has only {} representatives".format(key, value) for key, value in
                               class_counts.items()]
            if len(class_counts) == 1:
                list_of_classes = ": " + list_of_classes[0]
            else:
                n_lines_to_show = 5
                if len(list_of_classes) > n_lines_to_show:
                    suffix = '\n\t... {} more.'.format(len(list_of_classes) - n_lines_to_show)
                else:
                    suffix = ''
                list_of_classes = ". List of classes:\n\t" + ",\n\t".join(list_of_classes[:n_lines_to_show]) + suffix

            insights.append("{}{}".format(sentence, list_of_classes))

        return self._add_cls_name_prefix(insights)

    def _outlier_counts_insights(self):
        insights = []
        for feature_name, outliers_info in self.outliers_count.items():
            max_n_sigma = max(outliers_info.keys())

            n_sigma_info = outliers_info[max_n_sigma]
            n_sigma_range = "[{:.3g},{:.3g}]".format(
                n_sigma_info['mean'] - n_sigma_info['n_sigma'] * n_sigma_info['std'],
                n_sigma_info['mean'] + n_sigma_info['n_sigma'] * n_sigma_info['std'], )
            sentence = "Feature '{}' has {} outliers with respect to {} - a {} range." \
                .format(feature_name, n_sigma_info['n_outliers'], n_sigma_range, max_n_sigma)
            insights.append(sentence)

            log10_plus_one = numpy.log10(numpy.round(n_sigma_info['max'] + 1))

            if n_sigma_info['max'] >= 99 and log10_plus_one == int(log10_plus_one):
                insights.append("Feature '{}' has a maximal value of {} which doesn't seem to be a legit value".format(
                    feature_name, n_sigma_info['max']))

        return self._add_cls_name_prefix(insights)

    def insights(self):
        return list(itertools.chain(self.missing_values_report.insights(),
                                    self.covariance_shift_report.insights(),
                                    self._outlier_counts_insights(),
                                    self._class_count_insights()))
"""


import multiprocessing

import numpy
import sklearn
from matplotlib import pyplot as plt
from pytolemaic.utils.constants import FeatureTypes
from pytolemaic.utils.constants import REGRESSION, CLASSIFICATION
from pytolemaic.utils.dmd import DMD
from pytolemaic.utils.general import GeneralUtils
from pytolemaic.utils.label_encoder_wrapper import LabelEncoderProtected
from pytolemaic.utils.metrics import Metrics
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.metrics import brier_score_loss
from sklearn.pipeline import Pipeline
from pytolemaic.utils.general import get_logger

logger = get_logger(__name__)

class EncodeWrapper():
    def __init__(self, model, encode_target):
        self.model = model
        self.encode_target=encode_target

    def fit(self, x, y, feature_types=None, feature_names=None):
        self.perform_xencoding = feature_types is not None and FeatureTypes.categorical in feature_types
        if feature_types is None:
            logger.warning("Feature types are not available, skipping label encoding")

        if self.perform_xencoding:
            self.xencoder = LabelEncoderProtected()
            self.xencoder.fit(x, feature_types=feature_types, feature_names=feature_names)
        else:
            self.xencoder = None

        if self.encode_target:
            self.yencoder = LabelEncoderProtected()
            self.yencoder.fit(y, feature_types=[FeatureTypes.categorical], feature_names='target')
        else:
            self.yencoder = None

        self.model.fit(x if self.xencoder is None else self.xencoder.transform(x),
                       y if self.yencoder is None else self.yencoder.transform(y).ravel())

    def predict(self, x):
        if self.encode_target:
            raise NotImplementedError("Inverse transform not yet implemented")
        return self.model.predict(x if self.xencoder is None else self.xencoder.transform(x))

    def predict_proba(self, x):
        return self.model.predict_proba(x if self.xencoder is None else self.xencoder.transform(x))

    @property
    def steps(self):
        return self.model.steps

class UncertaintyModelBase():

    def __init__(self, model, uncertainty_method: str,
                 ptype=None,
                 supported_methods: list = None):
        self.model = model
        self.uncertainty_method = uncertainty_method if uncertainty_method != 'default' else supported_methods[0]
        self.dmd_supported = None
        self.is_classification = GeneralUtils.is_classification(model)
        self.uncertainty_analysis_output = None

        if (self.is_classification and ptype != CLASSIFICATION) or \
                (not self.is_classification and ptype == CLASSIFICATION):
            raise ValueError(
                "{} does not support {}".format(type(self), ptype))

        if self.uncertainty_method not in supported_methods:
            raise NotImplementedError(
                "Uncertainty method {} is not in supported methods={}".format(
                    self.uncertainty_method, supported_methods))

    def uncertainty(self, dmd: DMD):
        raise NotImplementedError("")

    def fit_uncertainty_model(self, dmd_test, **kwargs):
        raise NotImplementedError("")

    def fit(self, dmd_test: DMD, **kwargs):
        self.dmd_supported = GeneralUtils.dmd_supported(model=self.model,
                                                        dmd=dmd_test)

        self.fit_uncertainty_model(dmd_test, **kwargs)
        return self

    def predict(self, dmd: DMD):
        if self.dmd_supported:
            if not isinstance(dmd, DMD):
                dmd = DMD(x=dmd)
            return self.model.predict(dmd)
        else:
            if isinstance(dmd, DMD):
                x = dmd.values
            else:
                x = dmd
            return self.model.predict(x)

    def predict_proba(self, dmd: DMD):
        if self.dmd_supported:
            if not isinstance(dmd, DMD):
                raise ValueError("DMD supported but input is not dmd")
            return self.model.predict_proba(dmd)
        else:
            if isinstance(dmd, DMD):
                x = dmd.values
            else:
                x = dmd
            return self.model.predict_proba(x)

    def plot_calibration_curve(self):
        raise NotImplementedError

    @classmethod
    def _get_decision_paths(cls, values, tree, y, uncertainty_threshold, low):

        if low:
            vals = values[y.ravel() <= uncertainty_threshold]
        else:
            vals = values[y.ravel() >= uncertainty_threshold]

        # vals = vals[:min(100, len(vals)), :]
        # decision_path_nodes = [tree.decision_path(val.reshape(1,-1)).indices for val in vals]
        decision_path_nodes = tree.decision_path(vals)

        return decision_path_nodes

    @classmethod
    def _get_nodes(cls, tree, uncertainty_threshold, low, min_samples_in_node=100):
        # get candidates nodes by n_samples and value.
        ids = tree.tree_.n_node_samples >= min_samples_in_node
        values = tree.tree_.value.ravel()

        impurity = tree.tree_.impurity
        impurity[impurity < 0] = 1e100
        rmse = numpy.sqrt(impurity).ravel()
        rmse[tree.tree_.impurity < 0] = -1

        if low:
            ids = ids & (values <= uncertainty_threshold) & (values >= rmse * 2)
        else:
            ids = ids & (values >= uncertainty_threshold) & (values >= rmse * 2)

        if not any(ids):
            return numpy.array([])

        nodes = numpy.arange(tree.tree_.node_count)[ids]

        # prune nodes
        for i in reversed(nodes):  # high to low
            for children in [tree.tree_.children_left[i], tree.tree_.children_right[i]]:
                nodes[nodes == children] = -1
        nodes = nodes[nodes > 0]
        return nodes

    def uncertainty_analysis(self, dmd_train: DMD = None, dmd_test: DMD = None,
                             n_jobs=multiprocessing.cpu_count() - 1, max_depth=5, min_samples_in_area=0.05,
                             percentile=10) -> dict:
        """
        Provides an analysis of areas with low/high uncertainty. Can assist in understanding where the model is good, and where it's not.
        :param dmd_train:
        :param dmd_test:
        :param n_jobs:
        :param max_depth: Maximum number of conditions to define an area.
        :param min_samples_in_area: Minimal train+test samples in reported area.
        :param percentile: percentile over uncertainty to provide threshold for high and low uncertainties.
        :return:
        """
        if dmd_train is not None:
            dmd = dmd_train
            if dmd_test is not None:
                dmd.append(dmd_test)
        elif dmd_test is not None:
            dmd = dmd_test
        else:
            raise ValueError("Require either dmd_train or dmd_test or both")

        if min_samples_in_area < 1:
            min_samples_in_area = int(dmd.n_samples * min_samples_in_area)

        # train RF on uncertainty output (because it's not necessarity RF in uncertainty estimator)
        estimator = RandomForestRegressor(random_state=0, n_jobs=n_jobs, max_depth=max_depth, n_estimators=50,
                                          min_samples_split=max(min_samples_in_area//2, 2),
                                          max_features="sqrt")

        pipeline = GeneralUtils.simple_imputation_pipeline(
            estimator)  # todo better imputation scheme whenever a model is created by package
        pipeline = EncodeWrapper(pipeline, encode_target=False)

        y = self.uncertainty(dmd)
        pipeline.fit(dmd.values, y.ravel(), feature_types=dmd.feature_types, feature_names=dmd.feature_names)

        # extract RF from pipeline
        names, estimators = zip(*pipeline.steps)
        imputer = estimators[-2]
        estimator = estimators[-1]

        low_uncertainty = numpy.percentile(y, percentile)
        high_uncertainty = numpy.percentile(y, 100 - percentile)

        areas_of_extreme_uncertainty = {'low': [], 'high': []}
        for mode in ['low', 'high']:
            extreme_areas = []
            for tree in estimator.estimators_:
                uncertainty_threshold = low_uncertainty if mode == 'low' else high_uncertainty

                # get nodes of interest
                nodes = self._get_nodes(tree, uncertainty_threshold=uncertainty_threshold, low=mode == 'low',
                                        min_samples_in_node=min_samples_in_area)
                if len(nodes) == 0:
                    continue

                # get relevant decision paths of nodes
                node_paths = self._get_decision_paths(imputer.transform(dmd.values), tree, y,
                                                      uncertainty_threshold=uncertainty_threshold, low=mode == 'low')

                # for each node, get msg
                for node in nodes:
                    # search for appropriate path.
                    found = None
                    for node_path_csr in node_paths:
                        nodes_path = node_path_csr.indices
                        if node in nodes_path:
                            found = nodes_path
                            break

                    assert found is not None

                    # this can be moved elsewhere if needed
                    from pytolemaic.analysis_logic.prediction_analysis.decision_tree_report import DecisionTreeExplainer
                    conditions = DecisionTreeExplainer._list_of_conditions(decision_tree=tree,
                                                                           nodes=found,
                                                                           feature_names=dmd.feature_names)

                    mean = tree.tree_.value[node][0][0]
                    rmse = numpy.sqrt(tree.tree_.impurity[node])

                    msg = "Uncertainty( {} ) = {}.".format(" & ".join(conditions),
                                                           "{:.2g}+-{:.2g}".format(mean, rmse))
                    # extreme_areas.append(msg)
                    extreme_areas.append(dict(where=" & ".join(conditions),
                                              uncertainty="{:.2g}+-{:.1g}".format(mean, rmse),
                                              n_samples="{} out of {}".format(tree.tree_.n_node_samples[node],
                                                                              dmd.n_samples)
                                              ))

            areas_of_extreme_uncertainty[mode] = extreme_areas

        areas_of_extreme_uncertainty = {'Low uncertainty': areas_of_extreme_uncertainty['low'],
                                        'High uncertainty': areas_of_extreme_uncertainty['high']}
        # from pprint import pprint
        # pprint(areas_of_extreme_uncertainty)
        return areas_of_extreme_uncertainty

class UncertaintyModelRegressor(UncertaintyModelBase):

    def __init__(self, model, uncertainty_method='mae'):
        super(UncertaintyModelRegressor, self).__init__(
            model=model, uncertainty_method=uncertainty_method,
            ptype=REGRESSION, supported_methods=['mae', 'rmse', 'quantile'])
        self._n_bins = 10
        self.actual_error = None
        self.mean_predicted_error = None
        self._cal_curve_uncertainty = None

    def fit_uncertainty_model(self, dmd_test, n_jobs=multiprocessing.cpu_count() - 1,
                              metric=Metrics.r2, do_analysis=True, **kwargs):

        dmd_test, cal_curve_samples = dmd_test.split(ratio=0.1)

        if self.uncertainty_method in ['mae', 'default']:
            estimator = RandomForestRegressor(
                random_state=0, n_jobs=n_jobs,
                n_estimators=kwargs.pop('n_estimators', 250))
            self.uncertainty_model = GeneralUtils.simple_imputation_pipeline(
                estimator)

            self.uncertainty_model = EncodeWrapper(self.uncertainty_model, encode_target=False)
            yp = self.predict(dmd_test)
            self.uncertainty_model.fit(dmd_test.values,
                                       numpy.abs(
                                           dmd_test.target.ravel() - yp.ravel()),
                                       feature_types=dmd_test.feature_types, feature_names=dmd_test.feature_names
                                       )
        elif self.uncertainty_method in ['rmse']:
            estimator = RandomForestRegressor(
                random_state=0, n_jobs=n_jobs,
                n_estimators=kwargs.pop('n_estimators', 100))

            self.uncertainty_model = GeneralUtils.simple_imputation_pipeline(
                estimator)
            self.uncertainty_model = EncodeWrapper(self.uncertainty_model, encode_target=False)

            yp = self.predict(dmd_test)
            self.uncertainty_model.fit(dmd_test.values,
                                       (dmd_test.target.ravel() - yp.ravel()) ** 2,
                                       feature_types=dmd_test.feature_types, feature_names=dmd_test.feature_names)
        elif self.uncertainty_method in ['quantile']:
            if isinstance(self.model, Pipeline):

                names, estimators = zip(*self.model.steps)
                estimator = estimators[-1]
                if hasattr(estimator, 'estimators_'):
                    self.model.estimators_ = estimator.estimators_

            if not hasattr(self.model, 'estimators_'):
                raise ValueError(
                    "'quantile' method can work only for models with estimators_ attrbiute. Use method='mae' instead")

        else:
            raise NotImplementedError("Method {} is not implemented"
                                      .format(self.uncertainty_method))

        if do_analysis:
            self.uncertainty_analysis_output = self.uncertainty_analysis(dmd_train=None, dmd_test=dmd_test)

            # the following section is purely for plotting/analysis purposes

            ## calibration curve
            y_pred = self.predict(cal_curve_samples).ravel()
            y_true = cal_curve_samples.target.ravel()

            delta = numpy.abs(y_true - y_pred)
            uncertainty = self.uncertainty(cal_curve_samples).ravel()
            self._cal_curve_uncertainty = uncertainty

            bins = numpy.linspace(0., max(uncertainty) + 1e-8, self._n_bins + 1).ravel()
            binids = numpy.digitize(uncertainty, bins) - 1

            bin_sums = numpy.bincount(binids, weights=uncertainty, minlength=len(bins))
            bin_true = numpy.bincount(binids, weights=delta, minlength=len(bins))
            bin_total = numpy.bincount(binids, minlength=len(bins))

            nonzero = bin_total > 10
            self.actual_error = (bin_true[nonzero] / bin_total[nonzero])
            self.mean_predicted_error = (bin_sums[nonzero] / bin_total[nonzero])

            # calibration curve by metric

            performance = []
            uncertainty_levels_middle = []
            for ibin in range(len(bins) - 1):
                inds = binids == ibin
                if numpy.sum(inds) < 10:
                    continue

                subset_score = metric.function(y_true=y_true[inds], y_pred=y_pred[inds])
                performance.append(subset_score)
                uncertainty_levels_middle.append((bins[ibin] + bins[ibin + 1]) / 2)

            self._cal_curve_metric = {'uncertainty': uncertainty_levels_middle,
                                      'score': performance,
                                      'metric': metric.name}

    def uncertainty(self, dmd: DMD):
        if isinstance(dmd, DMD):
            x = dmd.values
        else:
            x = dmd

        if self.uncertainty_method in ['mae']:
            out = self.uncertainty_model.predict(x)
            return out.reshape(-1, 1)
        elif self.uncertainty_method in ['rmse']:
            out = numpy.sqrt(self.uncertainty_model.predict(x))
            return out.reshape(-1, 1)
        elif self.uncertainty_method in ['quantile']:

            percentile = 10

            ys = numpy.zeros((len(x), len(self.model.estimators_)))
            for i, pred in enumerate(self.model.estimators_):
                ys[:, i] = pred.predict(x)

            err_down = numpy.percentile(ys, percentile, axis=1)
            err_up = numpy.percentile(ys, 100 - percentile, axis=1)

            return ((err_up - err_down) / 2).reshape(-1, 1)

        else:
            raise NotImplementedError("Method {} is not implemented"
                                      .format(self.uncertainty_method))

    def plot_calibration_curve(self):
        fig = plt.figure(figsize=(10, 10))
        ax1 = plt.subplot2grid((3, 1), (0, 0), rowspan=2)
        ax2 = plt.subplot2grid((3, 1), (2, 0))

        ax1.plot([min(self.mean_predicted_error), max(self.mean_predicted_error)],
                 [min(self.mean_predicted_error), max(self.mean_predicted_error)], "k:", label="Perfectly calibrated")

        ax1.plot(self.mean_predicted_error, self.actual_error, "s-",
                 color='b')

        # todo: remove _cal_curve_uncertainty from self
        ax2.hist(self._cal_curve_uncertainty, range=(0, 1.),
                 bins=self._n_bins,
                 color='b',
                 histtype="step", lw=2)

        ax1.set_ylabel("Actual error")
        ax1.set_ylim([min(self.actual_error) * 0.95, max(self.actual_error) * 1.05])
        ax1.legend(loc="lower right")
        ax1.set_title("Calibartion Curve for method {} (Regression)".format(self.uncertainty_method))

        ax2.set_xlabel("Mean predicted error")
        ax2.set_ylabel("Count")
        ax2.legend(loc="upper center", ncol=2)

        # curve by metric
        plt.figure()
        plt.plot(self._cal_curve_metric['uncertainty'], self._cal_curve_metric['score'], '*-r')

        plt.xlabel("Uncertainty level ({})".format(self.uncertainty_method))
        plt.ylabel("{} score".format(self._cal_curve_metric['metric']))
        plt.title("{} score vs uncertainty level".format(self._cal_curve_metric['metric']))



class UncertaintyModelClassifier(UncertaintyModelBase):

    def __init__(self, model, uncertainty_method='confidence'):
        super(UncertaintyModelClassifier, self).__init__(model=model,
                                                         uncertainty_method=uncertainty_method,
                                                         ptype=CLASSIFICATION,
                                                         supported_methods=['confidence', 'probability'])

        self._brier_loss = -1
        self._fraction_of_positives = None
        self._mean_uncertainty = None
        self._cal_curve_uncertainty = None
        self._n_bins = 10

    def fit_uncertainty_model(self, dmd_test, n_jobs=multiprocessing.cpu_count() - 1,
                              metric=Metrics.recall, do_analysis=True,
                              **kwargs):

        if self.uncertainty_method in ['probability']:
            cal_curve_samples = dmd_test
            # no fit logic required

        elif self.uncertainty_method in ['confidence']:
            if do_analysis:
                dmd_test, cal_curve_samples = dmd_test.split(ratio=0.1)

            estimator = RandomForestClassifier(
                random_state=0, n_jobs=n_jobs, n_estimators=100)

            self.uncertainty_model = GeneralUtils.simple_imputation_pipeline(
                estimator)
            self.uncertainty_model = EncodeWrapper(self.uncertainty_model, encode_target=True)

            y_pred = self.predict(dmd_test)
            is_correct = numpy.array(y_pred.ravel() == dmd_test.target.ravel(),
                                     dtype=int)

            self.uncertainty_model.fit(dmd_test.values,
                                       is_correct.ravel(),
                                       feature_types=dmd_test.feature_types, feature_names=dmd_test.feature_names)

        else:
            raise NotImplementedError("Method {} is not implemented"
                                      .format(self.uncertainty_method))

        if do_analysis:
            self.uncertainty_analysis_output = self.uncertainty_analysis(dmd_train=None, dmd_test=dmd_test)

            # analysis for plots :

            # calibration curve

            y_pred = self.predict(cal_curve_samples).ravel()
            y_true = cal_curve_samples.target.ravel()
            uncertainty = self.uncertainty(cal_curve_samples)
            self._cal_curve_uncertainty = uncertainty

            self._fraction_of_positives, self._mean_uncertainty = sklearn.calibration.calibration_curve(
                y_true=y_pred == y_true,
                y_prob=uncertainty,
                normalize=True,
                n_bins=self._n_bins,
                strategy='uniform')

            sample_weight = None
            self._brier_loss = brier_score_loss(
                y_true=y_pred == y_true,
                y_prob=1 - uncertainty,
                sample_weight=sample_weight,
                pos_label=1)

            # calibration curve by metric

            uncertainty = uncertainty / max(uncertainty)

            bins = numpy.linspace(0., max(uncertainty) + 1e-8, 5 + 1)
            binids = numpy.digitize(uncertainty.flatten(), bins.ravel()) - 1

            performance = []
            uncertainty_levels_middle = []
            for ibin in range(len(bins) - 1):
                inds = binids == ibin
                if numpy.sum(inds) < 10:
                    continue

                subset_score = metric.function(y_true=y_true[inds], y_pred=y_pred[inds])
                performance.append(subset_score)
                uncertainty_levels_middle.append((bins[ibin] + bins[ibin + 1]) / 2)

            self._cal_curve_metric = {'uncertainty': uncertainty_levels_middle,
                                  'score': performance,
                                  'metric': metric.name}

    def uncertainty(self, dmd: DMD):

        if self.uncertainty_method in ['probability']:
            yproba = self.predict_proba(dmd)
            # break ties
            yproba += 1e-10 * numpy.random.RandomState(0).rand(*yproba.shape)
            max_probability = numpy.max(yproba, axis=1).reshape(-1, 1)
            delta = max_probability - yproba
            yproba[delta == 0] = 0

            # delta[numpy.sum(delta, axis=1)>=20,:] = 0 # i

            # ratio of 2nd highest probability to max_probability
            out = numpy.max(yproba, axis=1).reshape(-1, 1) / max_probability
            return GeneralUtils.f5(out).reshape(-1, 1)
        elif self.uncertainty_method in ['confidence']:
            if isinstance(dmd, DMD):
                x = dmd.values
            else:
                x = dmd
            # return the probability it's a mistake
            out = self.uncertainty_model.predict_proba(x)[:, 0]
            return GeneralUtils.f5(out).reshape(-1, 1)
        else:
            raise NotImplementedError("Method {} is not implemented"
                                      .format(self.uncertainty_method))

    def plot_calibration_curve(self):
        fig = plt.figure(figsize=(10, 10))
        ax1 = plt.subplot2grid((3, 1), (0, 0), rowspan=2)
        ax2 = plt.subplot2grid((3, 1), (2, 0))

        ax1.plot([0, 1], [1, 0.5], "k:", label="Perfectly calibrated")

        ax1.plot(self._mean_uncertainty, self._fraction_of_positives, "s-",
                 color='b',
                 label="brier loss=%1.3f" % self._brier_loss)

        # todo: remove y_proba from self
        ax2.hist(self._cal_curve_uncertainty, range=(0, 1.),
                 bins=self._n_bins,
                 color='b',
                 histtype="step", lw=2)

        ax1.set_ylabel("Fraction of correct predictions")
        ax1.set_ylim([-0.05, 1.05])
        ax1.legend(loc="lower right")
        ax1.set_title("Calibartion Curve for method {}".format(self.uncertainty_method))

        ax2.set_xlabel("Mean uncertainty")
        ax2.set_ylabel("Count")
        ax2.legend(loc="upper center", ncol=2)

        # curve by metric
        plt.figure()
        plt.plot(self._cal_curve_metric['uncertainty'], self._cal_curve_metric['score'], '*-r')

        plt.xlabel("Uncertainty level ({})".format(self.uncertainty_method))
        plt.ylabel("{} score".format(self._cal_curve_metric['metric']))
        plt.title("{} score vs uncertainty level".format(self._cal_curve_metric['metric']))



import copy

import numpy
from pytolemaic import FeatureTypes
from sklearn.preprocessing import LabelEncoder
from pytolemaic.utils.general import get_logger
import pandas as pd
logger = get_logger(__name__)

class LabelEncoderProtected():
    def __init__(self, nan_list=()):
        self.encoders = []
        self.categorical_encoding_dict = {}
        self.nan_list = nan_list
        self.jibrish_value = '!@#$%^&*()'

    def _replace_nan_with_jibbrish(self, x):

        for k in self.nan_list:
            x[x==k] = numpy.nan

        nan_mask = pd.DataFrame(x).isnull().values

        x[nan_mask] = self.jibrish_value

        return x, nan_mask

    def _replace_jibbrish_with_nan(self, x):
        x[x==self.jibrish_value] = numpy.nan
        return x


    def fit(self, x, feature_types, feature_names=None):
        self.encoders = []
        self.categorical_encoding_dict = {}

        if isinstance(x, pd.DataFrame) or isinstance(x, pd.Series):
            x = x.values

        if len(x.shape) == 1:
            x = x.reshape(-1,1)

        if feature_types is None:
            raise ValueError("Couldn't encode dmd w/o feature_types")

        if isinstance(feature_types, str):
            feature_types = [feature_types]
        if len(feature_types) != x.shape[1]:
            raise ValueError("Inconsistent # of features and # of feature_types")
        if feature_names is None:
            feature_names = numpy.arange(x.shape[1])

        self._feature_names = feature_names
        self._feature_types = feature_types

        x = copy.deepcopy(x).astype(object)
        x, nan_mask = self._replace_nan_with_jibbrish(x)

        for i, (name, feature_type) in enumerate(zip(self._feature_names, self._feature_types)):
            logger.info("Fit encoding for feature #{}:'{}'. Feature type is '{}'"
                         .format(i, name, feature_type))

            if feature_type == FeatureTypes.categorical:
                le = LabelEncoder().fit(x[:,i].astype(str))
                self.encoders.append(le)
                self.categorical_encoding_dict[name] = {i: cls_ for i, cls_ in enumerate(le.classes_)}
            else:
                self.encoders.append(None)

        return self

    def _encode_protected(self, le, series):
        try:
            return le.transform(series), None
        except ValueError as e:
            classes = le.classes_
            mask = numpy.array([x in classes for x in series])

            new_values = numpy.zeros(mask.shape)
            new_values[mask] = le.transform(series[mask])
            new_values[~mask] = -1  # never seen these values before - new class
            return new_values, e

    def transform(self, x):
        if not self.encoders:
            raise ValueError("Call fit() before transform()")

        if isinstance(x, pd.DataFrame) or isinstance(x, pd.Series):
            x = x.values

        if len(x.shape) == 1:
            x = x.reshape(1, -1)

        x = copy.deepcopy(x).astype(object)
        x, nan_mask = self._replace_nan_with_jibbrish(x)

        for i, (name, feature_type) in enumerate(zip(self._feature_names, self._feature_types)):
            if feature_type == FeatureTypes.categorical:
                logger.info("Transform feature #{}:'{}'".format(i, name))

                x[:,i], e = self._encode_protected(le=self.encoders[i],
                                                   series=x[:,i].astype(str))
                if e is not None:
                    logger.warning(e)

        x[nan_mask] = numpy.nan

        return x.astype(float)

if __name__ == '__main__':

    dmd = pd.DataFrame([list('fff'),
                 list('abc'),
                 list('ddd'),
                [numpy.nan]*3])
    lew = LabelEncoderProtected().fit(dmd, feature_types=[FeatureTypes.categorical] * 3)
    dmd_out = lew.transform(dmd)

    print(dmd_out.values)

    dmd = pd.DataFrame([list('fff'),
                        list('abc'),
                        list('hhh'),
                        [numpy.nan] * 3])
    dmd_out = lew.transform(dmd)



    print(dmd_out.values)

    dmd = pd.DataFrame([list('fff'),
                 list('abc'),
                 list('ddd'),
                [numpy.nan]*3])[0]
    lew = LabelEncoderProtected().fit(dmd, feature_types=[FeatureTypes.categorical] * 3)
    dmd_out = lew.transform(dmd)

    print(dmd_out.values)
